import subprocess
import os
import fcntl
from pprint import pprint
from pygdbmi import gdbmiparser
import sys
import select
from distutils.spawn import find_executable
from multiprocessing import Lock

PYTHON3 = sys.version_info.major == 3
GDB_TIMEOUT_SEC = 1
MUTEX_AQUIRE_WAIT_TIME_SEC = int(1)
unicode = str if PYTHON3 else unicode


class GdbController():
    """
    Run gdb as a subprocess. Send commands and receive structured output.
    Create new object, along with a gdb subprocess

    Args:
        gdb_path (str): Command to run in shell to spawn new gdb subprocess
        gdb_args (list): Arguments to pass to shell when spawning new gdb subprocess
        verbose (bool): Print verbose output if True
    Returns:
        New GdbController object
    """

    def __init__(self, gdb_path='gdb', gdb_args=['--nx', '--quiet', '--interpreter=mi2'], verbose=False):
        self.verbose = verbose
        self.mutex = Lock()
        self.abs_gdb_path = None  # abs path to gdb executable
        self.cmd = []  # the shell command to run gdb
        self._buffer = ''  # string buffer for unifinished gdb output

        if not gdb_path:
            raise ValueError('a valid path to gdb must be specified')
        else:
            abs_gdb_path = find_executable(gdb_path)
            if abs_gdb_path is None:
                raise ValueError('gdb executable could not be resolved from "%s"' % gdb_path)
            else:
                self.abs_gdb_path = abs_gdb_path

        self.cmd = [self.abs_gdb_path] + gdb_args

        if verbose:
            print('Launching gdb: "%s"' % ' '.join(self.cmd))

        # Use pipes to the standard streams
        # In UNIX a newline will typically only flush the buffer if stdout is a terminal.
        # If the output is being redirected to a file, a newline won't flush
        self.gdb_process = subprocess.Popen(self.cmd, shell=False, stdout=subprocess.PIPE, stdin=subprocess.PIPE, stderr=subprocess.PIPE)

        # Set the file status flag (F_SETFL) on the pipes to be non-blocking
        # so we can attempt to read from a pipe with no new data without locking
        # the program up
        fcntl.fcntl(self.gdb_process.stdout, fcntl.F_SETFL, os.O_NONBLOCK)
        fcntl.fcntl(self.gdb_process.stderr, fcntl.F_SETFL, os.O_NONBLOCK)

        # save file numbers for use later
        self.stdout_fileno = self.gdb_process.stdout.fileno()
        self.stderr_fileno = self.gdb_process.stderr.fileno()
        self.stdin_fileno = self.gdb_process.stdin.fileno()

        self.read_list =  [self.stdout_fileno, self.stderr_fileno]
        self.write_list = [self.stdin_fileno]

    def write(self, mi_cmd_to_write,
                                    timeout_sec=GDB_TIMEOUT_SEC,
                                    verbose=False,
                                    raise_error_on_timeout=True,
                                    read_response=True):
        """Write to gdb process. Block while parsing responses from gdb for a maximum of timeout_sec.

        A mutex is obtained before writing and released before returning

        Args:
            mi_cmd_to_write (str or list): String to write to gdb. If list, it is joined by newlines.
            timeout_sec (int): Maximum number of seconds to wait for response before exiting. Must be >= 0.
            verbose (bool): Be verbose in what is being written
            raise_error_on_timeout (bool): If read_response is True, raise error if no response is received
            read_response (bool): Block and read response. If there is a separate thread running,
            this can be false, and the reading thread read the output.
        Returns:
            List of parsed gdb responses if read_response is True, otherwise []
        Raises:
            ValueError if there is no gdb subprocess running
            ValueError if mi_cmd_to_write is not valid
        """
        if not self.gdb_process:
            raise ValueError('gdb process is not attached')
        elif timeout_sec < 0:
            print('warning: timeout_sec was negative, replacing with 0')
            timeout_sec = 0

        verbose = self.verbose or verbose

        # Ensure proper type of the mi command
        if type(mi_cmd_to_write) in [str, unicode]:
            pass
        elif type(mi_cmd_to_write) == list:
            mi_cmd_to_write = '\n'.join(mi_cmd_to_write)
        else:
            raise ValueError('The gdb mi command must a be str or list. Got ' + str(type(mi_cmd_to_write)))

        if verbose:
            print('\nwriting: %s' % mi_cmd_to_write)

        if not mi_cmd_to_write.endswith('\n'):
            mi_cmd_to_write_nl = mi_cmd_to_write + '\n'

        _, outputready, _ = select.select([], self.write_list, [], timeout_sec)
        for fileno in outputready:
            if fileno == self.stdin_fileno:
                # ready to write
                self.gdb_process.stdin.write(mi_cmd_to_write_nl.encode())
                # don't forget to flush for Python3, otherwise gdb won't realize there is data
                # to evaluate, and we won't get a response
                self.gdb_process.stdin.flush()
            else:
                print('developer error: got unexpected fileno %d, event %d' % (fileno, event))

        if read_response is True:
            return self.get_gdb_response(timeout_sec=timeout_sec, raise_error_on_timeout=raise_error_on_timeout, verbose=verbose)
        else:
            return []

    def get_gdb_response(self, timeout_sec=GDB_TIMEOUT_SEC, raise_error_on_timeout=True, verbose=False):
        """Get response from GDB, and block while doing so. If GDB does not have any response ready to be read
        by timeout_sec, a ValueError is raised.

        A lock (mutex) is obtained before reading and released before returning. If the lock cannot
        be obtained, no data is read and an empty list is returned.

        Args:
            timeout_sec (float): Time to wait for reponse. Must be >= 0.
            raise_error_on_timeout (bool): Whether an exception should be raised if no response was found
            after timeout_sec
            verbose (bool): If true, more output it printed

        Returns:
            List of parsed GDB responses, returned from gdbmiparser.parse_response, with the
            additional key 'stream' which is either 'stdout' or 'stderr'

        Raises:
            ValueError if response is not received within timeout_sec
            ValueError if select returned unexpected file number
            ValueError if there is no gdb subprocess running
        """

        # validate inputs
        if not self.gdb_process:
            raise ValueError('gdb process is not attached')
        elif timeout_sec < 0:
            print('warning: timeout_sec was negative, replacing with 0')
            timeout_sec = 0

        verbose = self.verbose or verbose

        self.mutex.acquire(MUTEX_AQUIRE_WAIT_TIME_SEC)

        retval = []
        events, _, _ = select.select(self.read_list, [], [], timeout_sec)
        for fileno in events:
            # new data is ready to read
            if fileno == self.stdout_fileno:
                self.gdb_process.stdout.flush()
                raw_output = self.gdb_process.stdout.read()
                stream = 'stdout'

            elif fileno == self.stderr_fileno:
                self.gdb_process.stderr.flush()
                raw_output = self.gdb_process.stderr.read()
                stream = 'stderr'

            else:
                self.mutex.release()
                raise ValueError('Developer error. Got unexpected file number %d' % fileno)

            stripped_raw_response_list = [x.strip() for x in raw_output.decode().split('\n')]

            raw_response_list = list(filter(lambda x: x, stripped_raw_response_list))
            raw_response_list, self._buffer = _buffer_incomplete_responses(raw_response_list, self._buffer)

            # parse each response from gdb and put into a list
            for response in raw_response_list:
                if gdbmiparser.response_is_finished(response):
                    pass
                else:
                    parsed_response = gdbmiparser.parse_response(response)
                    parsed_response['stream'] = stream

                    retval.append(parsed_response)
                    if verbose:
                        pprint(parsed_response)

        self.mutex.release()

        if not retval and raise_error_on_timeout:
            raise ValueError('Did not get response from gdb after %s seconds' % GDB_TIMEOUT_SEC)
        else:
            return retval

    def exit(self):
        """Terminate gdb process
        Returns: None"""
        if self.gdb_process:
            self.gdb_process.terminate()
        self.gdb_process = None
        return None


def _buffer_incomplete_responses(raw_response_list, buf):
    """It is possible poll() returns EPOLLIN before gdb finished writing its response. In that
    case, a partial mi response was read, which cannot be parsed into structured data.
    We want to ALWAYS parse complete mi records. To do this, we store a buffer of gdb's
    output if gdb did not tell us it finished.

    Args:
        raw_response_list: List of gdb responses
        buf (str): Buffered gdb response from the past. This is incomplete and needs to be prepended to
        gdb's next output.

    Returns:
        (buffered_response_list, buffer)
    """

    if buf:
        # We have a partial response in our buffer. Combine it with
        # this response to hopefully form a complete response.
        raw_response_list[0] = buf + raw_response_list[0]
        # Erase buffer since we put it back into the output to be parsed.
        # If it's still not complete, it will be set back in this buffer with the
        # new output.
        buf = ''

    num_responses = len(raw_response_list)
    for i, response in enumerate(raw_response_list):
        # i is zero indexed
        # num_responses is one indexed
        if response.startswith('^done'):
            # we got a response from gdb, but we need to make sure
            # the entire gdb completed writing it's response and there isn't some more
            # mi output that we need. Output is ready to be parsed ONLY when we get (gdb) in
            # the next response.
            if (i + 1) > (num_responses - 1):
                # This was the last response, but it's missing the "finished"
                # response! We got an incomplete result, which won't be parsed
                # correctly!

                # Store the partial response in a buffer and combine it with the next
                # response
                buf = response
                # don't process this incomplete response!
                raw_response_list = raw_response_list[1:-1]

    return (raw_response_list, buf)



import sys
import re
from setuptools import find_packages, setup, Command
from pygdbmi.tests import test_app

EXCLUDE_FROM_PACKAGES = []

readme = open('README.rst', 'r').read()

with open('pygdbmi/__init__.py', 'r') as fd:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                        fd.read(), re.MULTILINE).group(1)


class TestCommand (Command):
    description = 'test task'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        sys.exit(test_app.main())


setup(
    name='pygdbmi',
    version=version,
    author='Chad Smith',
    author_email='grassfedcode@gmail.com',
    description=('Parse gdb machine interface output with Python'),
    long_description=readme,
    url='https://github.com/cs01/pygdbmi',
    license='MIT',
    packages=find_packages(exclude=EXCLUDE_FROM_PACKAGES),
    include_package_data=True,
    keywords=['gdb', 'python', 'machine-interface', 'parse', 'frontend'],
    scripts=[],
    entry_points={},
    extras_require={},
    zip_safe=False,
    cmdclass={'test': TestCommand},
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
)



#!/usr/bin/env python3

"""
Unit tests

Run from top level directory: ./tests/test_app.py
"""

import os
import unittest
import subprocess
from pygdbmi.gdbmiparser import parse_response, assert_match
from pygdbmi.gdbcontroller import GdbController


class TestPyGdbMi(unittest.TestCase):

    def test_parser(self):
        """Test that the parser returns dictionaries from gdb mi strings as expected"""

        # Test basic types
        assert_match(parse_response('^done'), {'type': 'result', 'payload': None, 'message': 'done', "token": None})
        assert_match(parse_response('~"done"'), {'type': 'console', 'payload': 'done', 'message': None})
        assert_match(parse_response('@"done"'), {"type": 'target', "payload": 'done', 'message': None})
        assert_match(parse_response('&"done"'), {"type": 'log', "payload": 'done', 'message': None})
        assert_match(parse_response('done'), {'type': 'output', 'payload': 'done', 'message': None})

        # Test escape sequences
        assert_match(parse_response('~""'), {"type": "console", "payload": "", 'message': None})
        assert_match(parse_response('~"\b\f\n\r\t\""'), {"type": "console", "payload": '\b\f\n\r\t\"', 'message': None})
        assert_match(parse_response('@""'), {"type": "target", "payload": "", 'message': None})
        assert_match(parse_response('@"\b\f\n\r\t\""'), {"type": "target", "payload": '\b\f\n\r\t\"', 'message': None})
        assert_match(parse_response('&""'), {"type": "log", "payload": "", 'message': None})
        assert_match(parse_response('&"\b\f\n\r\t\""'), {"type": "log", "payload": '\b\f\n\r\t\"', 'message': None})

        # Test a real world Dictionary
        assert_match(parse_response('=breakpoint-modified,bkpt={number="1",empty_arr=[],type="breakpoint",disp="keep",enabled="y",addr="0x000000000040059c",func="main",file="hello.c",fullname="/home/git/pygdbmi/tests/sample_c_app/hello.c",line="9",thread-groups=["i1"],times="1",original-location="hello.c:9"}'),
                                    {'message': 'breakpoint-modified',
                                    'payload': {'bkpt': {'addr': '0x000000000040059c',
                                                         'disp': 'keep',
                                                         'enabled': 'y',
                                                         'file': 'hello.c',
                                                         'fullname': '/home/git/pygdbmi/tests/sample_c_app/hello.c',
                                                         'func': 'main',
                                                         'line': '9',
                                                         'number': '1',
                                                         'empty_arr': [],
                                                         'original-location': 'hello.c:9',
                                                         'thread-groups': ['i1'],
                                                         'times': '1',
                                                         'type': 'breakpoint'}},
                                     'type' : 'notify',
                                     'token': None})

        #Test records with token
        assert_match(parse_response('1342^done'), {'type': 'result', 'payload': None, 'message': 'done', "token": 1342})


    def test_controller(self):
        """Build a simple C program, then run it with GdbController and verify the output is parsed
        as expected"""
        SAMPLE_C_CODE_DIR = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'sample_c_app')
        SAMPLE_C_BINARY = os.path.join(SAMPLE_C_CODE_DIR, 'a.out')
        # Build C program
        subprocess.check_output(["make", "-C", SAMPLE_C_CODE_DIR, '--quiet'])

        # Initialize object that manages gdb subprocess
        gdbmi = GdbController()

        # Load the binary and its symbols in the gdb subprocess
        responses = gdbmi.write('-file-exec-and-symbols %s' % SAMPLE_C_BINARY, timeout_sec=2)

        # Verify output was parsed into a list of responses
        assert(len(responses) != 0)
        response = responses[0]
        assert(set(response.keys()) == set(['message', 'type', 'payload', 'stream', 'token']))
        assert(response['message'] == 'thread-group-added')
        assert(response['type'] == 'notify')
        assert(response['payload'] == {'id': 'i1'})
        assert(response['stream'] == 'stdout')
        assert(response['token'] == None)

        responses = gdbmi.write(['-file-list-exec-source-files', '-break-insert main'])
        assert(len(responses) != 0)

        # Close gdb subprocess
        responses = gdbmi.exit()
        assert(responses is None)
        assert(gdbmi.gdb_process is None)


def main():
    loader = unittest.TestLoader()
    suite = unittest.TestSuite()

    suite.addTests(loader.loadTestsFromTestCase(TestPyGdbMi))

    runner = unittest.TextTestRunner(verbosity=1)
    result = runner.run(suite)
    return len(result.errors) + len(result.failures)


if __name__ == '__main__':
    main()



"""
Python parser for gdb's machine interface interpreter.

Parses string output from gdb with the "--interpreter=mi2" flag into
structured objects.

See more at https://sourceware.org/gdb/onlinedocs/gdb/GDB_002fMI.html#GDB_002fMI

"""


import re
from pygdbmi.printcolor import print_cyan, print_red, print_green
from pprint import pprint


def parse_response(gdb_mi_text):
    """Parse gdb mi text and turn it into a dictionary.

    See https://sourceware.org/gdb/onlinedocs/gdb/GDB_002fMI-Stream-Records.html#GDB_002fMI-Stream-Records
    for details on types of gdb mi output.

    Args:
        gdb_mi_text (str): String output from gdb

    Returns:
        dict with the following keys:
        type (either 'notify', 'result', 'console', 'log', 'target', 'done'),
        message (str or None),
        payload (str, list, dict, or None)
    """
    if _GDB_MI_NOTIFY_RE.match(gdb_mi_text):
        token, message, payload = _get_notify_msg_and_payload(gdb_mi_text)
        return {'type': 'notify',
                'message': message,
                'payload': payload,
                'token'  : token}

    elif _GDB_MI_RESULT_RE.match(gdb_mi_text):
        token, message, payload = _get_result_msg_and_paylod(gdb_mi_text)
        return {'type': 'result',
                'message': message,
                'payload': payload,
                'token'  : token}

    elif _GDB_MI_CONSOLE_RE.match(gdb_mi_text):
        return {'type': 'console',
                'message': None,
                'payload': _GDB_MI_CONSOLE_RE.match(gdb_mi_text).groups()[0]}

    elif _GDB_MI_LOG_RE.match(gdb_mi_text):
        return {'type': 'log',
                'message': None,
                'payload': _GDB_MI_LOG_RE.match(gdb_mi_text).groups()[0]}

    elif _GDB_MI_TARGET_OUTPUT_RE.match(gdb_mi_text):
        return {'type': 'target',
                'message': None,
                'payload': _GDB_MI_TARGET_OUTPUT_RE.match(gdb_mi_text).groups()[0]}

    elif response_is_finished(gdb_mi_text):
        return {'type': 'done',
                'message': None,
                'payload': None}

    else:
        # This was not gdb mi output, so it must have just been printed by
        # the inferior program that's being debugged
        return {'type': 'output',
                'message': None,
                'payload': gdb_mi_text}


def response_is_finished(gdb_mi_text):
    """Return true if the gdb mi response is ending
    Returns: True if gdb response is finished"""
    if _GDB_MI_RESPONSE_FINISHED_RE.match(gdb_mi_text):
        return True
    else:
        return False


def assert_match(actual_char_or_str, expected_char_or_str):
    """If values don't match, print them and raise a ValueError, otherwise,
    continue
    Raises: ValueError if argumetns do not match"""
    if expected_char_or_str != actual_char_or_str:
        print('Expected')
        pprint(expected_char_or_str)
        print('')
        print ('Got')
        pprint(actual_char_or_str)
        raise ValueError()


# ========================================================================
# All functions and variables below are used internally to parse mi output
# ========================================================================


# Print text to console as it's being parsed to help debug
_DEBUG = False

# GDB machine interface output patterns to match
# https://sourceware.org/gdb/onlinedocs/gdb/GDB_002fMI-Stream-Records.html#GDB_002fMI-Stream-Records

# https://sourceware.org/gdb/onlinedocs/gdb/GDB_002fMI-Result-Records.html#GDB_002fMI-Result-Records
# In addition to a number of out-of-band notifications,
# the response to a gdb/mi command includes one of the following result indications:
# done, running, connected, error, exit
_GDB_MI_RESULT_RE = re.compile('^(\d*)\^(\S+?)(,(.*))?$')

# https://sourceware.org/gdb/onlinedocs/gdb/GDB_002fMI-Async-Records.html#GDB_002fMI-Async-Records
# Async records are used to notify the gdb/mi client of additional
# changes that have occurred. Those changes can either be a consequence
# of gdb/mi commands (e.g., a breakpoint modified) or a result of target activity
# (e.g., target stopped).
_GDB_MI_NOTIFY_RE = re.compile('^(\d*)[*=](\S+?),(.*)$')

# https://sourceware.org/gdb/onlinedocs/gdb/GDB_002fMI-Stream-Records.html#GDB_002fMI-Stream-Records
# "~" string-output
# The console output stream contains text that should be displayed
# in the CLI console window. It contains the textual responses to CLI commands.
_GDB_MI_CONSOLE_RE = re.compile('~"(.*)"', re.DOTALL)

# https://sourceware.org/gdb/onlinedocs/gdb/GDB_002fMI-Stream-Records.html#GDB_002fMI-Stream-Records
# "&" string-output
# The log stream contains debugging messages being produced by gdb's internals.
_GDB_MI_LOG_RE = re.compile('&"(.*)"', re.DOTALL)

# https://sourceware.org/gdb/onlinedocs/gdb/GDB_002fMI-Stream-Records.html#GDB_002fMI-Stream-Records
# "@" string-output
# The target output stream contains any textual output from the
# running target. This is only present when GDB's event loop is truly asynchronous,
# which is currently only the case for remote targets.
_GDB_MI_TARGET_OUTPUT_RE = re.compile('@"(.*)"', re.DOTALL)

# Response finished
_GDB_MI_RESPONSE_FINISHED_RE = re.compile('^\(gdb\)\s*$')

_WHITESPACE = [' ', '\t', '\r', '\n']

_GDB_MI_CHAR_DICT_START = '{'
_GDB_MI_CHAR_ARRAY_START = '['
_GDB_MI_CHAR_STRING_START = '"'
_GDB_MI_VALUE_START_CHARS = [_GDB_MI_CHAR_DICT_START, _GDB_MI_CHAR_ARRAY_START, _GDB_MI_CHAR_STRING_START]


def _get_notify_msg_and_payload(result):
    """Get notify message and payload dict"""
    groups = _GDB_MI_NOTIFY_RE.match(result).groups()
    token  = int(groups[0]) if groups[0] != '' else None
    message = groups[1].strip()
    result_of_status = groups[2].strip()
    _, payload = _parse_dict(result_of_status)
    return token, message, payload


def _get_result_msg_and_paylod(result):
    """Get result message and payload dict"""
    groups = _GDB_MI_RESULT_RE.match(result).groups()
    token  = int(groups[0]) if groups[0] != '' else None
    message = groups[1]
    if groups[2] is None:
        payload = None
    else:
        str_to_parse = groups[2].strip()
        i, payload = _parse_dict(str_to_parse)
    return token, message, payload


def _parse_dict(to_parse):
    """Parse dictionary, with optional starting character '{'
    return (tuple):
        Number of characters parsed from to_parse
        Parsed dictionary
    """
    if _DEBUG:
        print_red('obj')
        print_red(to_parse)
    i = 0
    obj = {}
    while i < len(to_parse):
        c = to_parse[i]
        if c in _WHITESPACE:
            pass
        elif c in ['{', ',']:
            pass
        elif c == '}':
            # end of object, exit loop
            break
        else:
            chars_used, key, val = _parse_key_val(to_parse[i:])
            i = i + chars_used
            obj[key] = val
        i += 1
    if _DEBUG:
        print_green(obj)
    return i, obj


def _parse_key_val(to_parse):
    """Parse key, value combination
    return (tuple):
        Number of characters parsed from to_parse
        Parsed key (string)
        Parsed value (either a string, array, or dict)
    """
    if _DEBUG:
        print_red('keyval')
        print_red(to_parse)

    i = 0
    size, key = _parse_key(to_parse[i:])
    i += size
    size, val = _parse_val(to_parse[i:])
    i += size
    if _DEBUG:
        print_green(key)
        print_green(val)
    return i, key, val


def _parse_key(to_parse):
    """Parse key, value combination
    return (tuple):
        Number of characters parsed from to_parse
        Parsed key (string)
    """
    if _DEBUG:
        print_red('key')
        print_red(to_parse)

    buf = ''
    i = 0
    key = ''
    while i < len(to_parse) - 1:
        c = to_parse[i]
        if c == '=':
            key = buf
            # consume '=' sign so caller doesn't get it again
            i += 1
            break
        buf += c
        i += 1
    if _DEBUG:
        print_green(key)
    return i, key


def _parse_val(to_parse):
    """Parse value from string
    return (tuple):
        Number of characters parsed from to_parse
        Parsed value (either a string, array, or dict)
    """
    if _DEBUG:
        print_red('val')
        print_red(to_parse)

    i = 0
    val = ''
    buf = ''
    while i < len(to_parse) - 1:
        c = to_parse[i]

        if c == '{':
            # Start object
            size, val = _parse_dict(to_parse[i:])
            i += size
            break
        elif c == '[':
            # Start of an array
            size, val = _parse_array(to_parse[i:])
            i += size
            break
        elif c == '"':
            # Start of a string
            size, val = _parse_str(to_parse[i:])
            i += size
            break
        else:
            buf += c
        i += 1

    if _DEBUG:
        print_green(val)
    return i, val


def _parse_array(to_parse):
    """Parse an array
    return (tuple):
        Number of characters parsed from to_parse
        Parsed array
    """
    if _DEBUG:
        print_red('array')
        print_red(to_parse)

    assert_match(_GDB_MI_CHAR_ARRAY_START, to_parse[0])

    # Skip first open bracket so we don't end up in an
    # endless loop trying to re-parse the array
    i = 1

    arr = []
    while i < len(to_parse) - 1:
        c = to_parse[i]

        if c in _GDB_MI_VALUE_START_CHARS:
            size, val = _parse_val(to_parse[i:])
            arr.append(val)
            i += size
        elif c in _WHITESPACE:
            pass
        elif c == ',':
            pass
        elif c == ']':
            # Stop when this array has finished. Note
            # that elements of this array can be also be arrays.
            break
        i += 1
    if _DEBUG:
        print_green(arr)
    return i, arr


def _parse_str(to_parse):
    """Parse a string
    return (tuple):
        Number of characters parsed from to_parse
        Parsed string, without surrounding quotes
    """

    # characters that gdb escapes that should not be
    # escaped by this parser
    CHARS_TO_REMOVE_GDB_ESCAPE = ['"']

    if _DEBUG:
        print_red('string')
        print_red(to_parse)

    assert_match(_GDB_MI_CHAR_STRING_START, to_parse[0])
    i = 1  # Skip the opening quote
    buf = ''
    while i < len(to_parse) - 1:
        c = to_parse[i]
        if _DEBUG:
            print_cyan(c)

        if c == '\\' and i < (len(to_parse) - 1):
            # We are on a backslash and there is another character after the backslash
            # to parse. Handle this case specially since gdb escaped it for us

            # Get the next char that is being escaped
            c2 = to_parse[i + 1]
            if c2 in CHARS_TO_REMOVE_GDB_ESCAPE:
                # only store the escaped character in the buffer; don't store the backslash
                # (don't leave it escaped)
                buf += to_parse[i + 1]
            else:
                # store the backslash and the following character in the buffer (leave it escaped)
                buf += c + to_parse[i + 1]

            # consume the backslash and the next character
            i += 2

        else:
            # Quote is closed. Exit (and don't include the end quote).
            if c == '"':
                break
            buf += c
            i += 1

    string = buf
    if _DEBUG:
        print_green(string)
    return i, string
